header {* trials with "~~/src/HOL/Library/Polynomial" *}
theory Test_Polynomial 
imports Complex_Main
        "~~/src/HOL/Library/Polynomial"
        "~~/src/HOL/Library/Polynomial" "~~/src/HOL/Library/Fraction_Field"
begin

section {* Sections from /Library/Polynomial.thy ===============================================*}
subsection {* TEST:Auxiliary: operations for lists (later) representing coefficients ...........*}
value "nth_default"
value "nth_default (222::int) [1,2,3] 0" --"1"
value "nth_default (222::int) [1,2,3] 1" --"2"
value "nth_default (222::int) [1,2,3] 2" --"3"
value "nth_default (222::int) [1,2,3] 3" --"222"
value "xxx" --""

value "cCons" --""
value "cCons 222 [1,2,3::int]" --"[222, 1, 2, 3] compare pCons below"
value "xxx" --""

subsection {* TEST:Almost everywhere zero functions ............................................*}
value "xxx" --""

subsection {* TEST:Definition of type @{text poly} .............................................*}
(*value "Poly" --"Abstraction violation:
                  constant Poly"          FH? *)
value "     [:1, 2, 3::int:]"   --"see subsection {* List-style syntax for polynomials"
value "poly [:1, 2, 3::int:] x" --"non-executable function"
value "poly [:1, 2, 3::int:] 1" --"value of the function"

value "Fract"
value "Fract 2 (3::int)"
(*value "[:Fract 2 (3::int):]" --"Type int fract not of sort {zero,equal}
                                  No type arity fract :: equal"
*)
value "Fract [:3::int, 2, 1:] [:1::int, 2:]"
value "Fract [:[:3::int, 2, 1:]:] [:[:1::int, 2:]:]"
value "xxx" --""

subsection {* TEST:Degree of a polynomial ......................................................*}
value "degree"
value "degree [:0::int:]"
value "degree [:0::int, 1, 2, 3:]"
value "degree [:0:]" --"List.gen_length 0 (coeffs (Poly (cCons 0 (coeffs (Poly []))))) - Suc 0"
value "xxx" --""

subsection {* TEST:The zero polynomial .........................................................*}
value "[:0:]"
value "HOL.eq [:0:] 0"      --"equal_class.equal 
                              (coeffs (Poly (cCons 0 (coeffs (Poly []))))) (coeffs (Poly []))"
value "HOL.eq [:0::int:] 0" --"True"

value "nth_default 0 (coeffs [:0:]) 0"
value "coeff [:0     :] 0"     --"nth_default 0 (coeffs (Poly [])) 0"
value "coeff [:0::int:] 0"     --"0"
value "coeff (0:: int poly) 0" --"0"
value "coeff (0:: int poly) 0" --"0"
value "coeff (0:: int poly) 1" --"0"
value "coeff (0:: int poly) 2" --"0"
value "coeff (0           ) 2" --"nth_default 0 (coeffs (Poly [])) (Suc (Suc 0))"
value "xxx" --""

subsection {* TEST:List-style constructor for polynomials ......................................*}
value "pCons" --""
value "pCons (1::int) [:2, 3::nat:]" --"see [::] syntax below"
value "xxx" --""

subsection {* TEST:List-style syntax for polynomials ...........................................*}
value "pCons"
value "pCons 1 [:2, 3::nat:]"  --"Poly [1, 2, 3]"
value "pCons 1 [:2, 3::int:]"  --"Poly [1, 2, 3]"
value "pCons 1 [:2, 3::real:]" --"Poly [1, 2, 3]"
value "pCons 1 [:2, 3::complex:]"

value "Fract 2 (3::int)"
(*value "pCons (Fract 2 (3::int))" --"Type int fract not of sort {zero,equal}
                                      No type arity fract :: equal"
*)

value "poly [:1, 2, 3::int:] 2" --""
value "xxx" --""
value "xxx" --""
value "xxx" --""

subsection {* TEST:Representation of polynomials by lists of coefficients ......................*}
value "coeffs"
value "coeffs [:1, 2, 3::int:]"

value "HOL.equal"
value "HOL.equal [:1, 2, 3::int:] [:1, 2, 3::int:]" --"True"

value "is_zero"
value "is_zero [:1, 2, 3::int:]"
value "is_zero [:   0::int:]"
value "xxx" --""

subsection {* TEST:Fold combinator for polynomials .............................................*}
value "fold" --"('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b"
value "fold (op +)"
value "fold (op +) [1,2,3::int]"
value "fold (op +) [1,2,3::int] 0" --"6"
ML {*
fun fold _ [] y = y
  | fold f (x :: xs) y = fold f xs (f x y);
*}
(*stepwise evaluation:*)
(*     fold f      (x :: xs     ) y       = ...*)
value "fold (op +) (1 # [2,3]   ) 0::int"                                        --"6"
(*                                    ... = fold f      xs    (f      x y     )*)
value "                                     fold (op +) [2,3] ((op +) 1 0::int)" --"6"
value "fold (op +) (2 # [3::int]) 1::int"                                        --"6"
value "                                     fold (op +) [3]   ((op +) 2 1::int)" --"6"
value "fold (op +) (3 # []      ) 3::int"                                        --"6"
value "                                     fold (op +) []    ((op +) 3 3::int)" --"6"
(*     fold _      []             y = y*)
value "fold (op +) []             6::int"                                        --"6"

value "fold_coeffs" --"('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a poly \<Rightarrow> 'b \<Rightarrow> 'b"
value "fold_coeffs (op +)"
value "fold_coeffs (op +) [:1,2,3::int:]"
value "fold_coeffs (op +) [:1,2,3::int:] 0" --"6"

text {* fold_coeffs is required in 
  # Horner Schema in section Canonical morphism
  # Multiplication
  # Synthetic division
  # Long division
  # Composition of polynomials
*}
value "xxx" --""

subsection {* TEST:Canonical morphism on polynomials -- evaluation .............................*}
value "fold_coeffs" --""
value "fold_coeffs (\<lambda>a f x. a + x * f x)" --""
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:]" --""
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:] (\<lambda>x. 0)"   -- {* The Horner Schema *}
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:] (\<lambda>x. 0) 0" --"1"
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:] (\<lambda>x. 0) 1" --"6"
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:] (\<lambda>x. 0) 2" --"= 12 
  = 1*(2^0) + 2*(2^1) + 3*(2^2)
  = 1*1     + 2*2     + 3*4"
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:] (\<lambda>x. 0) 3" --"= 34"
value "fold_coeffs (\<lambda>a f x. a + x * f x) [:1,2,3::int:] (\<lambda>x. 0) x" 
  --"= fold (\<lambda>u ua ub. u + ub * ua ub) 
    (rev (coeffs (Poly (1 # coeffs (Poly (2 # coeffs (Poly (3 # coeffs (Poly []))))))))) (\<lambda>u. 0) x
    :: int"
value "xxx" --""

subsection {* TEST:Monomials ...................................................................*}
value "coeff" --"'a poly \<Rightarrow> nat \<Rightarrow> 'a"
value "coeff [:1,2,3::int:] 1" --"2"

value "monom" --"'a \<Rightarrow> nat \<Rightarrow> 'a poly"
value "monom (3::int) 0" --"Poly [3]"    --"lemma monom_0: monom a 0 = pCons a 0 "
value "monom (3::int) 1" --"Poly [0, 3]"
value "monom (3::int) 2" --"Poly [0, 0, 3]" --"----vvv!!!"
value "monom (3::int) 3" --"Poly [0, 0, 0, 3]"

value "coeff (monom a m) n" 
  --"nth_default 0 (coeffs (Poly (if a = 0 then [] else replicate m 0 @ [a]))) n"
value "coeff (monom (3::int) 2) 0" --"0"
value "coeff (monom (3::int) 2) 1" --"0"
value "coeff (monom (3::int) 2) 2" --"3   !!!!!!!!!!!!!!!"
value "coeff (monom (3::int) 2) 3" --"0"

value "monom a (Suc n) = pCons 0 (monom a n)" --"lemma monom_Suc"
value "monom (3::int) (Suc 1)"     --"Poly [0, 0, 3]"
value "pCons 0 (monom (3::int) 1)" --"Poly [0, 0, 3]"

value "degree (monom (3::int) 2)" --"2"
value "degree (monom (0::int) 2)" --"0"

(*lemma fold_coeffs_monom: equality of *functions*:
value "fold_coeffs f (monom a n) = f 0 ^^ n \<circ> f a" --"Wellsortedness error:
                                                      Type 'a \<Rightarrow> 'a not of sort equal"
*)
value "op ^^" --""
value "op ^^ [11::int]" --""
value "op ^^ [11::int] 1" --""

value "(op +)" --""
value "fold_coeffs" --""
value "fold_coeffs (op +)" --""
value "fold_coeffs (op +) (monom (3::int) 2)"   --"_"
value "fold_coeffs (op +) (monom (3::int) 2) 0" --"3"
value "xxx" --""

subsection {* TEST:Addition and subtraction ....................................................*}
value "coeff ([:1,1,1::int:] + [:1,2,3::int:]) 0"       --"2"
value "coeff ([:1,1,1::int:] + [:1,2,3::int:]) 1"       --"3"
value "coeff ([:1,1,1::int:] + [:1,2,3::int:]) 2"       --"4"
value "coeff ([:1,1,1::int:] + [:1,2,3::int:]) 3"       --"0"

value "coeff [:1,1,1::int:] 0 + coeff [:1,2,3::int:] 0" --"2"
value "coeff [:1,1,1::int:] 1 + coeff [:1,2,3::int:] 1" --"3"
value "coeff [:1,1,1::int:] 2 + coeff [:1,2,3::int:] 2" --"4"
value "coeff [:1,1,1::int:] 3 + coeff [:1,2,3::int:] 3" --"0"

value " [:1,1,1::int:] + [:1,2,3::int:]"   --"Poly [2, 3, 4]"
value "-[:1,1,1::int:] + [:1,2,3::int:]"   --"Poly [0, 1, 2]"
value " [:1,1,1::int:] - [:1,2,3::int:]"   --"Poly [0, -1, -2]"

(*fun plus_coeffs*)
value "plus_coeffs ((1::int) # [1,1]) (1 # [2,3])"  --"[2, 3, 4]"
(*
value "((1::int) + 1) ## (plus_coeffs [1,1] [2,3])" --"[2, 3, 4]" ?!? ?FH?*)

(*lemma poly_add [simp]: "poly (p + q) x = poly p x + poly q x"*)
value "poly ([:1,1,1::int:] + [:1,2,3::int:]) x" --"non-executable function?"
value "poly ([:1,1,1::int:] + [:1,2,3::int:]) 1" --"9
                            2*1^0 + 3*1^1 + 4*1^2 = 9"
value "xxx" --""

subsection {* TEST:Multiplication by a constant, polynomial multiplication, unit polynomial ....*}
value "smult (2::int)"                --"scalar multiplication"
value "smult (2::int) [:1,2,3::int:]" --"Poly [2, 4, 6]"

(*use definition "p * q = fold_coeffs (\<lambda>a p. smult a q + pCons 0 p) p 0" !!!!!!!!!!!!!!!*)
value " [:1,2::int:] * [:3,4::int:]"       --"Poly [3, 10, 8]"
value " [:1,2,1::int:] * [:1,1::int:]"     --"Poly [1, 3, 3, 1]"
value " [:1,2,1::int:] * [:1,3,3,1::int:]" --"Poly [1, 5, 10, 10, 5, 1]"

(*decompose definition "p * q = fold_coeffs (\<lambda>a p. smult a q + pCons 0 p) p 0"*)
(*definition "fold_coeffs f p i = foldr f (coeffs p) i"*)
(*List.thy: 
primrec foldr :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b" where
foldr_Nil:  "foldr f [] = id" |
foldr_Cons: "foldr f (x # xs) = f x \<circ> foldr f xs"
---------------------------------------------------extended to all arguments:
            "foldr f []       y = y" |
            "foldr f (x # xs) y = (f x \<circ> foldr f xs) y"
*)
(*stepwise evaluation ONLY of standardexample for foldr:*)
(*     foldr f      (x :: xs     ) y       = ...*)
value "foldr (op +) (1 # [2,3]   ) 0::int"                              --"6"
(*     ... = (f      x \<circ> foldr f      xs   ) y*)
value "      ((op +) 1 \<circ> foldr (op +) [2,3]) 0::int"                    --"6"
(*                  ... (f      x \<circ> foldr f      xs ) y*)
value "      ((op +) 1 \<circ> (op +) 2 \<circ> foldr (op +) [3]) 0::int"           --"6"
(*                             ... (f      x \<circ> foldr f      xs) y*)
value "      ((op +) 1 \<circ> (op +) 2 \<circ> (op +) 3 \<circ> foldr (op +) []) 0::int" --"6"
(*                                             foldr _     [] = id*)
value "      ((op +) 1 \<circ> (op +) 2 \<circ> (op +) 3 \<circ> id) 0::int"              --"6"

(*stepwise evaluation "p * q = fold_coeffs (\<lambda>a p. smult a q + pCons 0 p) p 0":*)
(*    "p            * q = ...*)
value "[:1,2::int:] * [:3,4::int:]"                                             --"Poly [3, 10, 8]"
(*.. = fold_coeffs (\<lambda>a p. smult a q + pCons 0 p) p 0":*)
value "fold_coeffs (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) [:1,2::int:] (0::int poly)" --"......."
(*     fold_coeffs f                                        p             i = ...
 ... = foldr f                                        (coeffs p           ) i*)
value "foldr (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) (coeffs [:1,2::int:]) (0::int poly)" --"...."
(*     foldr f                                        (       x # xs      ) i = ...
 ... = (f                                        x 
          \<circ> foldr f                                         xs  ) i*)
value "((\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) 1 
\<circ> foldr (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) [2]) 0"                        --"Poly [3, 10, 8]"
(*     (f                                        x 
            \<circ> foldr f                                         xs ) i = ...
 ... = (f x \<circ> f x 
                \<circ> foldr f                                        []) i*)
value "((\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) 1 
            \<circ> (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) 2
                \<circ> foldr (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) []) 0"         --"Poly [3, 10, 8]"
                                                                                --"--------------"
value " (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) 1 0"                           --"Poly [3, 4]"
value "       (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) 2 0"               --"?0?:  Poly [0, 6, 8]"
value "           foldr (\<lambda>a p. smult a [:3,4::int:] + pCons 0 p) [] 0"          --"Poly []"
value "              smult 2 [:3,4::int:]"                    --"Poly [6, 8]"
value "                                     pCons (0::int) 0" --"Poly []"
value "              smult 2 [:3,4::int:] + pCons (0::int) 0" --"Poly [6, 8]"

value "                                     pCons (0::int) [:3,4::int:]" --"?0?:  Poly [0, 3, 4]"
value "xxx" --""

subsection {* TEST:Lemmas about divisibility ...................................................*}
value "xxx" --""

subsection {* TEST:Polynomials form an integral domain .........................................*}
value "xxx" --""
                               
subsection {* TEST:Polynomials form an ordered integral domain .................................*}
value "xxx" --""

subsection {* TEST:Synthetic division and polynomial roots .....................................*}
value "synthetic_divmod"
value "synthetic_divmod [:1,3,3,1::int:] 0" --"(Poly [3, 3, 1], 1)"
value "synthetic_divmod [:1,3,3,1::int:] 1" --"(Poly [7, 4, 1], 8)"
value "synthetic_divmod [:1,3,3,1::int:] 2" --"(Poly [13, 5, 1], 27)"
value "synthetic_divmod [:1,3,3,1::int:] 3" --"(Poly [21, 6, 1], 64)"
value "synthetic_divmod [:1,3,3,1::int:] 4" --"(Poly [31, 7, 1], 125)"
value "synthetic_divmod [:1,3,3,1::int:] 5" --"(Poly [43, 8, 1], 216)"
value "xxx" --""
value "xxx" --""

value "synthetic_div" --""
value "synthetic_div [:1,3,3,1::int:] 0" --"Poly [3, 3, 1]"
value "synthetic_div [:1,3,3,1::int:] 1" --"Poly [7, 4, 1]"
value "synthetic_div [:1,3,3,1::int:] 2" --"Poly [13, 5, 1]"
value "synthetic_div [:1,3,3,1::int:] 3" --"Poly [21, 6, 1]"
value "synthetic_div [:1,3,3,1::int:] 4" --"Poly [31, 7, 1]"
value "synthetic_div [:1,3,3,1::int:] 5" --"Poly [43, 8, 1]"
value "xxx" --""
value "xxx" --""

subsection {* TEST:Long division of polynomials ................................................*}
value "xxx" --""

subsection {* TEST:Order of polynomial roots ...................................................*}
value "xxx" --""

subsection {* TEST:GCD of polynomials ..........................................................*}
value "xxx" --""

subsection {* TEST:Composition of polynomials ..................................................*}
value "xxx" --""

section {* Additional trials ===================================================================*}

value "xxx" --""
value "xxx" --""
value "xxx" --""

end
