header {* experiments on scanning and parsing
          following Stefan Berghofers presentation
          at the Isabelle Developer Workshop 2010 *}

theory Test_Parsers
imports Main
begin

text {* parse following Stefan Berghofer at Isabelle Developer Workshop 2010 *}
ML {* 
fun filtered_input str =
  filter Token.is_proper (Outer_Syntax.scan Position.none str);
fun parse p str = Scan.finite Token.stopper p (filtered_input str);
*}

section {* scanner combinators from src/Pure/General/scan.ML *}
text {* ATTENTION, there are several similar definitions:
src> grep -r 'infix ' * | grep '\-\- ' | less
HOL/Import/proof_kernel.ML:    infix 7 |-- --|
HOL/Import/proof_kernel.ML:    infix 5 :-- -- ^^
HOL/Import/mono_scan.ML:infix 7 |-- --|
HOL/Import/mono_scan.ML:infix 5 :-- -- ^^
HOL/Import/scan.ML:infix 7 |-- --|
HOL/Import/scan.ML:infix 5 :-- -- ^^
Pure/General/scan.ML:infix 5 -- :-- :|-- |-- --| ^^;
*}

subsection {* one typical usage per combinator *}
text {* typical signatures *}
ML {*
parse;
(*fn : (Token.T list -> 'a * Token.T list) -> string -> 'a * Token.T list*)
      Parse.nat;
      (*fn : int Parse.parser*)
      (*fn : (Token.T list -> 'a * Token.T list)*)
parse Parse.nat;
(*fn : string -> int * Token.T list*)

parse Parse.nat "123";
(*(123, []) : int * Token.T list*)
*}

subsubsection {* op >> *}
ML {*
Parse.nat;
(fn i => i + 1);
op >>;
(Parse.nat >> (fn i => i + 1));
parse;
parse (Parse.nat >> (fn i => i + 1));
(*
parse (Parse.nat >> (fn i => i + 1));
       int Parse.parser
                     int -> int
                 ('a -> 'b * 'c) * ('b -> 'd) -> 'a -> 'd * 'c
      Token.T list -> int * Token.T list
(Token.T list -> 'a * Token.T list) -> string -> 'a * Token.T list
val it = fn : string -> int * Token.T list
*)
*}

subsubsection {* op || *}
ML {*
parse (Args.$$$ "aaa" || Args.$$$ "bbb" || Args.$$$ "ccc") "aaa 111";
parse (Args.$$$ "aaa" || Args.$$$ "bbb" || Args.$$$ "ccc") "ccc 111";
(*val it = ("ccc", [Token (("111", ({}, {})), (Nat, "111"), Slot)])
   : string * Token.T list*)
parse ((Args.$$$ "aaa" -- Parse.nat) || (Args.$$$ "bbb" -- Parse.nat)) "aaa 111";
parse ((Args.$$$ "aaa" -- Parse.nat) || (Args.$$$ "bbb" -- Parse.nat)) "bbb 222";
*}

text {* || requires arguments of equal type; 2 tricks that help: *}
ML {* 
(* (1) works for prefixes complete enough to trigger a semantic action *)
fun exec1 str = "EXEC IMMEDIATELY: " ^ str;
fun exec2 (str, int) = "EXEC IMMEDIATELY: " ^ str ^ ", " ^ string_of_int int;

(* (2) works for combinations of complete prefixes *)
fun drop_prefix _ = "PREFIX DROPPED";

Args.$$$ "aaa" >> exec1; 
Args.$$$ "bbb" -- Parse.nat >> exec2;

Args.$$$ "aaa" >> exec1       || Args.$$$ "bbb" -- Parse.nat >> exec2;
Args.$$$ "aaa" >> drop_prefix || Args.$$$ "bbb" -- Parse.nat >> drop_prefix;
*}


subsubsection {* op :-- *}
ML {* (*from src/Pure/Isar/*)
            Parse.sym_ident : Token.T list -> string * Token.T list;
            (*fn : string Parse.parser*)
                                (fn "-" => Scan.succeed () | _ => Scan.fail);
                                (fn "-" => Scan.succeed ());
                                (*fn : string -> 'a -> unit * 'a*)
            Parse.sym_ident :-- (fn "-" => Scan.succeed () | _ => Scan.fail);
            (*fn : Token.T list -> (string * unit) * Token.T list*)
op >>;
(*fn : ('a -> 'b * 'c) * ('b -> 'd) -> 'a -> 'd * 'c*)

val minus = Parse.sym_ident :-- (fn "-" => Scan.succeed () | _ => Scan.fail) >> #1;
(*fn : Token.T list -> string * Token.T list*)
"------------------------------------------";
parse minus "-aaa";
parse minus "- -";
(*WHY NOT SIMPLY Args.$$$ "-" ?*)
(*parse minus "111 -";*)
"------------------------------------------";
infix 5 :--;
fun (scan1 -- scan2) = scan1 :-- (fn _ => scan2);
*}

subsubsection {* op -- *}
ML {*
parse (minus -- Parse.nat)  "- 111";

Args.$$$ "theory";
Args.name;
(Args.$$$ "theory" -- Args.name);
parse (Args.$$$ "theory" -- Args.name);
(*
parse (Args.$$$ "theory")                GOON ?!?!?
  "theory";

parse (Args.$$$ "theory")
  "theory Sketch imports Main";
parse (Args.$$$ "theory" -- Args.name)
  "theory Sketch imports Main";
parse (Args.$$$ "theory" -- Args.name -- Args.$$$ "imports" -- Args.name)
  "theory Sketch imports Main";
*)
*}

subsubsection {* op :|-- TODO*}
ML {*
fun token atom = Scan.ahead Parse.not_eof --| atom;
(*fn : (Token.T list -> 'a * 'b) -> Token.T list -> Token.T * 'b*)

val ident = token
  (Parse.short_ident || Parse.long_ident || Parse.sym_ident || Parse.term_var ||
    Parse.type_ident || Parse.type_var || Parse.number);
(*fn : Token.T list -> Token.T * Token.T list*)

(ident >> Token.content_of) :|-- (fn x =>
  (case Lexicon.read_variable x of SOME v => Scan.succeed v | NONE => Scan.fail));
(*fn : Token.T list -> indexname * Token.T list*)
*}

subsubsection {* op |-- *}
ML {* (*from Pure/Isar/args.ML, fun attribs*)
parse (Parse.$$$ "[" |-- Parse.list1 Parse.name) "[111, bbb, ccc]";
*}

subsubsection {* op --| *}
ML {* (*from Pure/Isar/args.ML, fun attribs*)
parse (Parse.list1 Parse.name --| Parse.$$$ "]") " 111, bbb, ccc]";
*}

section {* parse compound suffixes *}
subsection {* Lists *}
ML {* (*from Pure/Isar/args.ML, fun attribs*)
Parse.$$$ "[" |-- Parse.list1 Parse.name --| Parse.$$$ "]";
(*fn : Token.T list -> bstring list * Token.T list*)

parse (Parse.$$$ "[" |-- Parse.list Parse.name --| Parse.$$$ "]") "[aaa, bbb, ccc]"; 
*}

subsection {* terms *}
text {* see src/Pure/isac/smltest/Pure/Isar/Test_Parse_Term.thy *}

subsection {* option and repetition *}
ML {*
val xxx = (Args.$$$ "top" -- (Scan.option (Args.$$$ "equiv"))
       -- Args.$$$ "dots");
parse xxx "top dots";
parse xxx "top equiv dots";

val yyy = Scan.repeat1 (Args.$$$ "equiv");
parse yyy "equiv equiv equiv";
parse yyy "equiv";
(*
parse yyy "";
*** exception FAIL (SOME "keyword expected,\nbut end-of-file was found") raised (line 135 of "General/scan.ML")*)
*}

subsection {* recursive parsing *}
ML {*
fun body s xs = (Args.$$$ "top" >> drop_prefix 
             || (Args.$$$ "bullet" -- (body s)) >> drop_prefix) xs;
parse (body "top") "top";
parse (body "top") "bullet top";
parse (body "top") "bullet bullet bullet top";
"........ parsed all, however: ..........";

parse (body "top") "top bullet top";
(*val it =
   ("PREFIX DROPPED",
    [Token (("bullet", ({}, {})), (Ident, "bullet"), Slot),
     Token (("top", ({}, {})), (..., "top"), Slot)])
   : string * Token.T list*)
"-------- fst argument seems irrelavant ? ...";
parse (body "") "top";
parse (body "") "bullet top";
parse (body "") "bullet bullet bullet top";
parse (body "") "top bullet top";
*}

section {* GOON *}
subsection {* GOON *}


end